from collections import deque


class SQLToken:
    def __init__(self, lexer, ttype, value=""):
        self.line = lexer.line
        self.col = lexer.col
        self.ttype = ttype
        self.value = value


class SQLLexer:

    WHITESPACE = "WHITESPACE"
    COMMENT = "COMMENT"
    NAME = "NAME"
    QUOTED = "QUOTED"
    INT = "INT"
    FLOAT = "FLOAT"
    CHAR = "CHAR"
    COMMA = ","
    DOT = "."
    EOS = ";"

    def __init__(self, sql, whitespaces=True):
        self.itr = iter(sql)
        self.buffer = deque()
        self.line = 1
        self.col = 1
        self.whitespaces = whitespaces

    def load_buffer(self):
        while len(self.buffer) < 3:
            c = next(self.itr, None)
            self.buffer.append(c)
        return self.buffer[0]

    def read_char(self):
        self.load_buffer()
        c = self.buffer.popleft()
        if not c:
            return None
        self.col += 1
        if c == "\n":
            self.line += 1
            self.col = 1
        return c

    def push(self, c):
        self.buffer.appendleft(c)
        self.col -= 1

    def is_whitespace(self, c):
        return c and c in (" ", "\t", "\n")

    def is_alpha(self, c):
        return c and ((c >= "a" and c <= "z") or (c >= "A" and c <= "Z") or c == "_")

    def is_digit(self, c):
        return c and c >= "0" and c <= "9"

    def read_whitespace(self):
        x = SQLToken(self, self.WHITESPACE)
        c = self.read_char()
        while self.is_whitespace(c):
            x.value += c
            c = self.read_char()
        self.push(c)
        return x

    def read_name(self):
        x = SQLToken(self, self.NAME)
        c = self.read_char()
        while self.is_alpha(c) or self.is_digit(c):
            x.value += c
            c = self.read_char()
        self.push(c)
        return x

    def read_number(self):
        x = SQLToken(self, self.INT)
        c = self.read_char()
        while self.is_digit(c):
            x.value += c
            c = self.read_char()
        if c != ".":
            self.push(c)
            return x
        x.ttype = self.FLOAT
        x.value += c
        c = self.read_char()
        while self.is_digit(c):
            x.value += c
            c = self.read_char()
        self.push(c)
        return x

    def read_comment_single(self):
        x = SQLToken(self, self.COMMENT)
        c = self.read_char()
        while c and c != "\n":
            x.value += c
            c = self.read_char()
        self.push(c)
        return x

    def read_comment_multi(self):
        x = SQLToken(self, self.COMMENT)
        c = self.read_char()
        while c and x.value[-2:] != "*/":
            x.value += c
            c = self.read_char()
        self.push(c)
        return x

    def read_quoted(self):
        q = self.read_char()
        x = SQLToken(self, self.CHAR if q == "'" else self.QUOTED)
        while True:
            c = self.read_char()
            if c != q:
                x.value += c
            elif c == q:
                c1 = self.read_char()
                if c1 == q:
                    x.value += q
                    c = self.read_char()
                else:
                    self.push(c1)
                    break
        return x

    def __iter__(self):
        return self

    def __next__(self):
        self.load_buffer()
        if self.buffer[0] is None:
            raise StopIteration()
        if self.is_whitespace(self.buffer[0]):
            token = self.read_whitespace()
            if self.whitespaces:
                return token
        if self.buffer[0] == "#":
            return self.read_comment_single()
        if (self.buffer[0], self.buffer[1]) == ("-", "-"):
            return self.read_comment_single()
        if (self.buffer[0], self.buffer[1]) == ("/", "*"):
            return self.read_comment_multi()
        if self.is_alpha(self.buffer[0]):
            return self.read_name()
        if self.buffer[0] == "[":
            self.read_char()
            token = self.read_name()
            if self.buffer[0] != "]":
                raise ValueError(f"{self.line}: {self.buffer[0]}")
            self.read_char()
            return token
        if self.is_digit(self.buffer[0]):
            return self.read_number()
        if self.buffer[0] in ("'", '"', "`"):
            return self.read_quoted()
        if self.buffer[0] in ("(", ")"):
            token = SQLToken(self, self.buffer[0], self.buffer[0])
            self.read_char()
            return token
        if (self.buffer[0], self.buffer[1]) in (
            ("<", "="),
            (">", "="),
            ("!", "="),
            ("<", ">"),
            ("+", "="),
        ):
            v = self.buffer[0] + self.buffer[1]
            token = SQLToken(self, v, v)
            self.read_char()
            self.read_char()
            return token
        if self.buffer[0] in (".", ",", ";", "+", "-", "/", "*", "<", ">", "=", "|"):
            v = self.buffer[0]
            token = SQLToken(self, v, v)
            self.read_char()
            return token
        raise ValueError(f"{self.line}: {self.buffer[0]}")



import MySQLdb
from etlx.dbi.abc import DBI
from etlx.sql import SQL

class SQL_MySQL(SQL):
    
    def quoted(self, *args):
        for i, name in enumerate(args):
            self.sql("," if i else "")
            self.sql('`' + name.replace('`', '``') + '`')
        return self


class DBI_MySQL_MySQLdb(DBI):

    SQL = SQL_MySQL

    def connect_factory(self, **kwargs):
        return MySQLdb.connect(**kwargs)

    



from etlx.abc.row import RowDict
from etlx.abc.mapper import RowMapper, ExtractAttr, ExtractItem, ExtractConst, Datetime
from etlx.abc.column import Column
from etlx.abc.index import Index
from etlx.abc.table import Table
from etlx.abc.metadata import MetadataMixIn



import psycopg2
from etlx.dbi.abc import DBI
from etlx.sql import SQL

class SQL_Postgres(SQL):
    pass
   

class DBI_Postgres_psycopg2(DBI):

    SQL = SQL_Postgres

    def connect_factory(self, **kwargs):
        return psycopg2.connect(**kwargs)




from .column import Column


class Index:

    def __init__(self, columns, *, name=None, table=None, metadata=None):
        self._name = name
        self._table = table
        self._columns = columns
        self._metadata = metadata

    @property
    def name(self):
        return self._name

    @property
    def columns(self):
        return self._columns

    @property
    def primary(self):
        return None

    @property
    def unique(self):
        return None

    @property
    def table(self):
        return self._table

    @property
    def metadata(self):
        return self._metadata or {}

    def __getattr__(self, name):
        if self._metadata is None or name not in self._metadata:
            raise AttributeError('no metadata available')
        return self._metadata[name]



import sqlite3
from etlx.dbi.abc import DBI
from etlx.sql import SQL

class SQLite_Cursor(sqlite3.Cursor):

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


class SQL_SQLite(SQL):
    pass
   

class DBI_SQLite_sqlite3(DBI):

    SQL = SQL_SQLite

    def connect_factory(self, **kwargs):
        return sqlite3.connect(**kwargs)

    def cursor(self, serverside=False):
        return self._dbapi.cursor(factory=SQLite_Cursor)



from collections import OrderedDict
from .column import Column
from .index import Index


class Table:

    def __init__(self, name, *, dbi=None, metadata=None):
        self._name = name
        self._dbi = dbi
        self._metadata = metadata
        self._columns = None
        self._indexes = None
        self._primaryIndex = None

    @property
    def name(self):
        return self._name

    @property
    def dbi(self):
        return self._dbi

    @property
    def metadata(self):
        return self._metadata or {}

    @property
    def columns(self):
        if self._columns is None:
            self._columns = self.readColumns()
        return self._columns

    @property
    def indexes(self):
        if self._indexes is None:
            self._indexes = self.readIndexes()
        return self._indexes

    @property
    def primaryIndex(self):
        if self._indexes is None:
            self._indexes = self.readIndexes()
        return self._primaryIndex

    @property
    def tableType(self):
        return self.metadata.get('TABLE_TYPE')

    @property
    def comment(self):
        return self.metadata.get('TABLE_COMMENT')

    def readColumns(self):
        result = OrderedDict()
        if self.dbi and hasattr(self.dbi, 'queryColumns'):
            for column in self.dbi.queryColumns(table=self):
                column._parent = self
                result[column.name] = column
        result.__setitem__ = None
        return result

    def readIndexes(self):
        result = OrderedDict()
        if self.dbi and hasattr(self.dbi, 'queryIndexes'):
            for index in self.dbi.queryIndexes(table=self):
                index._parent = self
                result[index.name] = index
                if index.primary:
                    self._primaryIndex = index
        result.__setitem__ = None
        return result



from collections import OrderedDict
from datetime import datetime
from etlx.abc import RowDict


class RowMapper(OrderedDict):

    def __call__(self, row):
        # result = map(lambda x: x(row), self.values())
        return RowDict((k, f(row)) for k, f in self.items())


class ExtractAttr:

    def __init__(self, key):
        self.key = key

    def __call__(self, row):
        return getattr(row, self.key, None)


class ExtractItem:

    def __init__(self, key):
        self.key = key

    def __call__(self, row):
        if isinstance(row, (list,tuple)):
            return row[self.key]
        return row.get(self.key, None)


class ExtractConst:

    def __init__(self, value):
        self.value = value

    def __call__(self, row):
        return self.value


class Datetime:

    def __init__(self, extractor):
        self.extractor = extractor

    def __call__(self, row):
        value = self.extractor(row)
        if isinstance(value, str):
            value = value.split('.')[0]
            value = datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
        return value



from io import StringIO
from decimal import Decimal
from datetime import datetime, date, time


class SQL:
    def __init__(self, dbi=None):
        self.dbi = dbi
        self.buffer = StringIO()

    def __len__(self):
        return self.buffer.tell()

    def __str__(self):
        return self.buffer.getvalue()

    def __bool__(self):
        return len(self) > 0

    def sql(self, *args):
        for sql in args:
            if not isinstance(sql, str):
                sql = str(sql)
            self.buffer.write(sql)
        return self

    def execute(self, *args, **kwargs):
        return self.dbi.execute(self, *args, **kwargs)

    def query(self, *args, **kwargs):
        return self.dbi.query(self, *args, **kwargs)

    def quoted(self, *args):
        for i, name in enumerate(args):
            self.sql("," if i else "")
            self.sql('"' + name.replace('"', '""') + '"')
        return self

    def arg(self):
        return self.sql("%s")

    def kwarg(self, name):
        return self.sql(f"%({name})s")

    def literal(self, *args):
        for i, value in enumerate(args):
            self.sql("," if i else "")
            if value is None:
                self.sql("NULL")
            elif isinstance(value, bool):
                self.sql("1" if value else "0")
            elif isinstance(value, (int, float, Decimal)):
                self.sql(str(value))
            elif isinstance(value, (datetime, date, time)):
                self.sql("'", str(value), "'")
            elif isinstance(value, str):
                value = value.replace("'", "''")
                value = value.replace("%", "%%")
                self.sql("'", value, "'")
            elif isinstance(value, tuple):
                self.sql("(").literal(*value).sql(")")
            else:
                raise NotImplementedError()
        return self

    def _list(self, func, iterable, separator=","):
        for i, x in enumerate(iterable):
            self.sql(separator if i else "")
            func(x)

    def SELECT(self, *args, **kwargs):
        self.sql("SELECT ")
        if not args:
            self.sql("*")
        else:
            self._list(self.quoted, args)
        return self

    def FROM(self, table, database=None):
        self.sql(" FROM ")
        if database:
            self.quoted(database).sql(".")
        self.quoted(table)
        return self

    def INSERT(self, table, **kwargs):
        self.sql("INSERT INTO ").quoted(table)
        self.sql(" (").quoted(*kwargs.keys()).sql(") VALUES (").literal(
            *kwargs.values()
        ).sql(")")
        return self

    def INSERT_CV(self, table, columns, *rows):
        self.sql("INSERT INTO ").quoted(table)
        self.sql(" (").quoted(*columns).sql(") VALUES ")
        for i, values in enumerate(rows):
            self.sql("," if i else "")
            self.sql("(").literal(*values).sql(")")
        return self

    def UPDATE(self, table, **kwargs):
        self.sql("UPDATE ").quoted(table).sql(" SET ")
        for i, (k, v) in enumerate(kwargs.items()):
            self.sql("," if i else "")
            self.quoted(k).sql("=").literal(v)
        return self

    def UPDATE_CV(self, table, columns, values):
        self.sql("UPDATE ").quoted(table).sql(" SET ")
        for i, (k, v) in enumerate(zip(columns, values)):
            self.sql("," if i else "")
            self.quoted(k).sql("=").literal(v)
        return self

    def DELETE(self, table):
        self.sql("DELETE FROM ").quoted(table)
        return self

    def WHERE(self, **kwargs):
        self.sql(" WHERE ")
        for i, (k, v) in enumerate(kwargs.items()):
            self.sql(" AND " if i else "")
            self.quoted(k).sql("=").literal(v)
        return self

    def WHERE_CV(self, columns, *keys):
        self.sql(" WHERE ")
        if isinstance(columns, str):
            self.quoted(columns).sql(" IN (")
            for i, value in enumerate(keys):
                self.sql("," if i else "")
                self.literal(value)
            self.sql(")")
        else:
            for i, values in enumerate(keys):
                self.sql(" OR (" if i else "(")
                for j, (k, v) in enumerate(zip(columns, values)):
                    self.sql(" AND " if j else "")
                    self.quoted(k).sql("=").literal(v)
                self.sql(")")
        return self



from etlx.abc.row import RowDict
from etlx.sql import SQL


class DBI:

    SQL = SQL

    def __init__(self, *, connect=None):
        self._dbapi = None
        self._connect_kwargs = dict()
        if isinstance(connect,dict):
            self._connect_kwargs.update(connect)
        self._close_on_exit = False

    @property
    def database(self):
        return self._connect_kwargs.get("database")

    @property
    def sql(self):
        return self.SQL(self)

    @property
    def connected(self):
        return self._dbapi is not None

    def connect_factory(self, **kwargs):
        raise NotImplementedError()

    def connect(self, **kwargs):
        params = dict()
        params.update(self._connect_kwargs)
        params.update(kwargs)
        params = filter(lambda x: (x[1] != None), params.items())
        params = dict(params)
        self._dbapi = self.connect_factory(**params)

    def close(self):
        if self.connected:
            self._dbapi.close()
        self._dbapi = None

    def commit(self):
        self._dbapi.commit()

    def rollback(self):
        if self._dbapi:
            self._dbapi.rollback()

    def cursor(self, serverside=False):
        return self._dbapi.cursor()

    def execute(self, _sql, *args, **kwargs):
        if not isinstance(_sql, str):
            _sql = str(_sql)
        with self.cursor(serverside=False) as cursor:
            cursor.execute(_sql, args or kwargs)
            return cursor

    def query(self, _sql, *args, **kwargs):
        if not isinstance(_sql, str):
            _sql = str(_sql)
        with self.cursor(serverside=True) as cursor:
            cursor.execute(_sql, args or kwargs)
            for row in cursor:
                yield RowDict((d[0], v) for d, v in zip(cursor.description, row))

    def readone(self, _sql, *args, **kwargs):
        if not isinstance(_sql, str):
            _sql = str(_sql)
        with self.cursor(serverside=True) as cursor:
            cursor.execute(_sql, args or kwargs)
            row = cursor.fetchone()
            if row:
                row = RowDict((d[0], v) for d, v in zip(cursor.description, row))
            return row

    def __enter__(self):
        if not self.connected:
            self.connect()
            self._close_on_exit = True
        else:
            self._close_on_exit = False
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.connected:
            return
        if exc_type:
            self.rollback()
        else:
            self.commit()
        if self._close_on_exit:
            self.close()



from collections import OrderedDict


class RowDict(OrderedDict):

    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def update(self, *args, **kwargs):
        for seq in args:
            if isinstance(seq, dict):
                seq = seq.items()
            for key, value in seq:
                self[key] = value
        for key, value in kwargs.items():
            self[key] = value

    def __getattr__(self, key):
        if key not in self:
            raise AttributeError(key)
        return self[key]


def rowExtract(x, columns):
    return RowDict((cname, x.get(cname)) for cname in columns)


def rowTuple(x, columns):
    return tuple(x.get(cname) for cname in columns)



from collections import OrderedDict
from itertools import groupby
from etlx.abc import Column, Index, MetadataMixIn, Table


class MySQL_Table(Table):

    @property
    def database(self):
        return self.metadata.get('TABLE_SCHEMA')


class MySQL_Column(Column):

    @property
    def database(self):
        return self.metadata.get('TABLE_SCHEMA')


class MySQL_Index(Index):

    @property
    def database(self):
        return self.metadata.get('TABLE_SCHEMA')

    @property
    def primary(self):
        return self.name == 'PRIMARY'

    @property
    def unique(self):
        return self.metadata.get('NON_UNIQUE') == 0


class MySQL_MetadataMixIn(MetadataMixIn):

    Table = MySQL_Table
    Column = MySQL_Column
    Index = MySQL_Index

    def queryTables(self, database=None):
        sql = "SELECT * FROM information_schema.TABLES WHERE"
        if database:
            sql += f" TABLE_SCHEMA='{database}'"
        else:
            sql += " TABLE_SCHEMA=DATABASE()"
        sql += " ORDER BY TABLE_NAME"
        result = []
        for row in self.query(sql):
            tableName = row['TABLE_NAME']
            table = self.Table(name=tableName, dbi=self, metadata=row)
            result.append(table)
        return result

    def queryColumns(self, *, table=None, database=None):
        sql = "SELECT * FROM information_schema.COLUMNS WHERE "
        if table is not None:
            sql += f" TABLE_NAME='{table.name}' AND TABLE_SCHEMA='{table.database}'"
        sql += " ORDER BY TABLE_NAME, ORDINAL_POSITION"
        result = []
        for row in self.query(sql):
            tableName = row['TABLE_NAME']
            columnName = row['COLUMN_NAME']
            column = self.Column(name=columnName, table=tableName, metadata=row)
            result.append(column)
        return result

    def queryIndexes(self, *, table=None, database=None):
        sql = "SELECT * FROM information_schema.STATISTICS WHERE"
        if table is not None:
            sql += f" TABLE_NAME='{table.name}' AND TABLE_SCHEMA='{table.database}' AND INDEX_SCHEMA='{table.database}'"
        sql += " ORDER BY TABLE_NAME, INDEX_NAME, SEQ_IN_INDEX"
        result = []
        iterable = self.query(sql)
        groupKey = lambda x: (x['TABLE_NAME'], x['INDEX_NAME'])
        for (TABLE_NAME, INDEX_NAME), group in groupby(iterable, groupKey):
            idxRow = None
            columns = OrderedDict()
            for row in group:
                if idxRow is None:
                    keys = ('TABLE_SCHEMA', 'TABLE_NAME', 'INDEX_NAME', 'NON_UNIQUE', 'INDEX_TYPE', 'INDEX_COMMENT')
                    idxRow = OrderedDict((k, row.get(k)) for k in keys)
                COLUMN_NAME = row['COLUMN_NAME']
                columns[COLUMN_NAME] = row
            columns.__setitem__ = None
            index = self.Index(name=INDEX_NAME, columns=columns, table=TABLE_NAME, metadata=idxRow)
            result.append(index)
        return result



from typing import Mapping
from collections import OrderedDict
from .table import Table, Column, Index


class MetadataMixIn:

    METADATA_LAZY_READ = True

    Table = Table

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tables = None
        self._tablesLazyRead = self.METADATA_LAZY_READ

    @property
    def tables(self) -> Mapping[str, Table]:
        if self._tables is None:
            self.loadTables()
        return self._tables

    def __call__(self, name) -> Table:
        return self.tables[name]

    def loadTables(self):
        self._tables = self.readTables(lazyRead=self._tablesLazyRead)

    def readTables(self, *, lazyRead=True):
        tables = OrderedDict()
        # all TABLES
        for table in self.queryTables():
            tables[table.name] = table
        tables.__setitem__ = None
        if lazyRead:
            return tables
        # init columns & indexes dicts
        for table in tables.values():
            table._columns = OrderedDict()
            table._indexes = OrderedDict()
        # all COLUMNS
        for column in self.queryColumns():
            table = tables[column.table]
            column._parent = table
            table._columns[column.name] = column
        # all INDEXES
        for index in self.queryIndexes():
            table = tables[index.table]
            index._parent = table
            table._indexes[index.name] = index
            if index.primary:
                table._primaryIndex = index
        # make immutable
        for table in tables.values():
            table._columns.__setitem__ = None
            table._indexes.__setitem__ = None
        return tables

    def queryColumns(self, *, table=None, TABLE_CATALOG=None, TABLE_SCHEMA=None):
        if isinstance(table, Table):
            table = table.name
        kwargs = dict(TABLE_NAME=table, TABLE_CATALOG=TABLE_CATALOG, TABLE_SCHEMA=TABLE_SCHEMA)
        sql = "SELECT * FROM information_schema.COLUMNS"
        sql += self._whereFilter(**kwargs)
        sql += " ORDER BY TABLE_NAME, ORDINAL_POSITION"
        result = []
        for row in self.queryAll(sql, **kwargs):
            row = self._rowUpperColumnNames(row)
            tableName = row['TABLE_NAME']
            columnName = row['COLUMN_NAME']
            column = self.Column(name=columnName, table=tableName, metadata=row)
            result.append(column)
        return result

#    def queryIndexes(self, table=None, TABLE_CATALOG=None, TABLE_SCHEMA=None):
#        return []



#!/usr/bin/python3

import os
from distutils.core import Command
import pkg_resources
import setuptools

BASEDIR = os.path.dirname(__file__)

with open(os.path.join(BASEDIR, "README.md"), "r") as f:
    README = f.read()
with open(os.path.join(BASEDIR, "VERSION"), "r") as f:
    VERSION = f.read().strip()
with open(os.path.join(BASEDIR, "etlx", "build.py"), "w") as f:
    GITHUB_WORKFLOW = os.environ.get("GITHUB_WORKFLOW")
    if GITHUB_WORKFLOW == "etlx-release":
        pass
    elif GITHUB_WORKFLOW == "etlx-build":
        GITHUB_RUN_NUMBER = os.environ.get("GITHUB_RUN_NUMBER")
        VERSION += f".dev{GITHUB_RUN_NUMBER}"
    else:
        CI_BUILD_ID = os.environ.get("CI_BUILD_ID", 0)
        VERSION += f".dev{CI_BUILD_ID}"
    f.write(f'__version__ = "{VERSION}"\n')


class Requirements(Command):

    description = "prints project requirements"
    user_options = [
        ("all", None, f"Output (default: console)"),
    ]

    def initialize_options(self):
        self.all = False

    def finalize_options(self):
        pass

    def run(self):
        requires = []
        install_requires = getattr(self.distribution, "install_requires", [])
        requires.extend(pkg_resources.parse_requirements(install_requires))
        extras_require = getattr(self.distribution, "extras_require", {}) if self.all else {}
        for extra, reqs in extras_require.items():
            marker = None
#            if not extra.startswith(":"):
#                marker = extra[1:]
#                e = pkg_resources.invalid_marker(marker)
#                if e:
#                    raise e
            for r in pkg_resources.parse_requirements(reqs):
                assert not r.marker
                r.marker = marker
                requires.append(r)
        result = "\n".join(sorted(map(str, requires)))
        print(result)


setuptools.setup(
    name="etlx",
    version=VERSION,
    author="Alexander Keda",
    author_email="kedikx.io@gmail.com",
    description="ETL & Co",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/kedikx/etlx",
    packages=setuptools.find_packages(exclude=("etlx_tests", "etlx_tests.*")),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["PyYAML"],
    extras_require={
        "mysql": ["mysqlclient"],
        "postgres": ["psycopg2"]
    },
    cmdclass={
        "requirements": Requirements,
    },
)



from typing import Mapping
import yaml
from copy import deepcopy
from importlib import import_module
from collections import OrderedDict


class ServiceFactory:

    def __init__(self, configFile=None):
        self.configFile = configFile
        self._catalog = None

    @property
    def catalog(self) -> Mapping:
        if self._catalog is None:
            self.load(self.configFile)
        return self._catalog

    def load(self, config):
        if self._catalog is None:
            self._catalog = OrderedDict()
        if not config:
            return
        config = loadYAML(config)
        for name, kwargs in config.items():
            if isinstance(kwargs, str):
                self.add(name, factory=kwargs)
            else:
                self.add(name, **kwargs)

    def add(self, name: str, factory=None, **kwargs):
        """Adds/replaces service entry"""
        if self._catalog is None:
            self._catalog = OrderedDict()
        if isinstance(factory, str):
            entry = self._catalog.get(factory, None)
            if entry:
                factory = entry[0]
                kwargs = self.mergeArgs(entry[1], **kwargs)
            else:
                func = self.findCallable(factory)
                if not callable(func):
                    raise ValueError(f"can't resolve {factory} into callable")
                factory = func
        self._catalog[name] = (factory, kwargs)

    def mergeArgs(self, base, **kwargs):
        """Compose result kwargs from base + optional kwags"""
        result = deepcopy(base)
        # adjust params
        for k, v in kwargs.items():
            c = result.get(k, None)
            if isinstance(c, dict) and isinstance(v, dict):
                c.update(v)
            else:
                result[k] = v
        return result

    def findCallable(self, factory: str):
        """Resolves factory name into callable: module.function"""
        s = factory.split('.')
        moduleName = '.'.join(s[:-1])
        funcName = s[-1:][0]
        if not moduleName or not funcName:
            return None
        module = import_module(name=moduleName)
        return getattr(module, funcName)

    def __getitem__(self, name):
        entry = self.catalog[name]
        return entry[1]

    def __getattr__(self, name):
        entry = self.catalog[name]

        def factory(**kwargs):
            kwargs = self.mergeArgs(entry[1], **kwargs)
            return entry[0](**kwargs)
        return factory


def loadYAML(filePath) -> Mapping:
    """Loads yaml as OrderedDict"""
    class Loader(yaml.SafeLoader):
        pass

    def constructMapping(loader, node):
        loader.flatten_mapping(node)
        return OrderedDict(loader.construct_pairs(node))
    Loader.add_constructor(yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, constructMapping)
    with open(filePath) as file:
        return yaml.load(file, Loader)
